import os
import time
import base64
import hashlib
import threading

import maried
from maried.core import AlreadyInQueueError, Denied

from mirte.core import Module
from sarah.event import Event
from sarah._itertools import iter_by_n
from joyce.base import JoyceChannel


class MariedChannelClass(JoyceChannel):
	def __init__(self, server, *args, **kwargs):
		super(MariedChannelClass, self).__init__(*args, **kwargs)
                self.server = server
                self.user = None
                self.login_token = None
                self.send_message({
                        'type': 'welcome',
                        'protocols': [0],
                        'note': 'The API is not stable'})

        def handle_stream(self, stream):
                if self.user is None:
                        stream.close()
                        self.send_message({
                                'type': 'error_upload',
                                'message': 'Please log in before uploading'})
                        return
                self.l.info('Download started')
                mf = self.server.desk.add_media(stream, self.user)
                self.l.info('Download finished: ' + repr(mf))

	def handle_message(self, data):
                if data['type'] == 'follow':
                        for followed in data['which']:
                                self.server._register_follower(self, followed)
                elif data['type'] == 'unfollow':
                        for followed in data['which']:
                                self.server._unregister_follower(self, followed)
                elif data['type'] == 'request_login_token':
                        self.login_token = base64.b64encode(os.urandom(6))
                        self.send_message({
                                'type': 'login_token',
                                'login_token': self.login_token})
                elif data['type'] == 'regenerate_accessKey':
                        if self.user is None:
                                self.send_message({
                                        'type': 'error_regenerate_accessKey',
                                        'message': 'not logged in'})
                                return
                        self.user.regenerate_accessKey()
                        self.user.save()
                        self.send_message({
                                'type': 'accessKey',
                                'accessKey': self.user.accessKey})
                elif data['type'] == 'login' or \
                     data['type'] == 'login_accessKey':
                        if ('username' not in data or
                            'hash' not in data):
                                self.send_message({
                                        'type': 'error_'+data['type'],
                                        'message': 'Expected user and hash'})
                                return
                        try:
                                user = self.server.desk.user_by_key(
                                                data['username'])
                        except KeyError:
                                self.send_message({
                                        'type': 'error_'+data['type'],
                                        'message': 'User does not exist'})
                                return
                        secret = user.passwordHash if data['type'] == 'login' \
                                        else user.accessKey
                        if secret is None:
                                self.send_message({
                                        'type': 'error_'+data['type'],
                                        'message': 'Secret not set'})
                                return
                        expected_hash = hashlib.md5(secret +
                                                self.login_token).hexdigest()
                        if expected_hash != data['hash']:
                                self.send_message({
                                        'type': 'error_'+data['type'],
                                        'message': 'Wrong password'})
                                return
                        self.user = user
                        if user.accessKey is None:
                                user.regenerate_accessKey()
                                user.save()
                        self.send_message({
                                'type': 'logged_in',
                                'accessKey': user.accessKey})
                elif data['type'] == 'request':
                        if self.user is None:
                                self.send_message({
                                        'type': 'error_request',
                                        'message': 'Please log in before '+
                                                        'requesting'})
                                return
                        try:
                                m = self.server.desk.media_by_key(
                                                data['mediaKey'])
                        except KeyError:
                                self.send_message({
                                        'type': 'error_request',
                                        'message': 'No such media'})
                                return
                        try:
                                m = self.server.desk.request_media(m, self.user)
                        except AlreadyInQueueError:
                                self.send_message({
                                        'type': 'error_request',
                                        'message': 'Already queued'})
                        except Denied:
                                self.send_message({
                                        'type': 'error_request',
                                        'message': 'Request denied'})
                elif data['type'] == 'cancel_request':
                        if not 'key' in data:
                                self.send_message({
                                        'type': 'error_cancel_request',
                                        'message': 'Missing key'})
                                return
                        try:
                                req = self.server.requests_ns.by_key(
                                                data['key'])
                        except KeyError:
                                self.send_message({
                                        'type': 'error_cancel_request',
                                        'message': 'Request not found'})
                                return
                        self.server.desk.cancel_request(req, self.user)
                elif data['type'] == 'move_request':
                        if not 'key' in data or not 'amount' in data:
                                self.send_message({
                                        'type': 'error_move_request',
                                        'message': 'Missing key or amount'})
                                return
                        try:
                                req = self.server.requests_ns.by_key(
                                        data['key'])
                        except KeyError:
                                self.send_message({
                                        'type': 'error_move_request',
                                        'message': 'Request not found'})
                                return
                        self.server.desk.move_request(req, int(data['amount'],
                                                        self.user))
                elif data['type'] == 'list_media':
                        self.server._send_all_media((self,))
                elif data['type'] == 'skip_playing':
                        self.server.desk.skip_playing(self.user)
                else:
                        self.send_message({
                                'type': 'error',
                                'message': 'unknown msg type %s' % \
                                                data['type']})

        def after_close(self):
                self.l.debug("Closed")
                self.server._remove_follower(self)

class JoyceRS(Module):
	def __init__(self, *args, **kwargs):
		super(JoyceRS, self).__init__(*args, **kwargs)
		self.joyceServer.channel_class = self._channel_constructor
		self.desk.on_media_changed.register(
				self._on_media_changed)
		self.desk.on_playing_changed.register(
				self._on_playing_changed)
		self.desk.on_requests_changed.register(
				self._on_requests_changed)
                self.following_lut = {
                                'requests': (set(), self._send_all_requests),
                                'playing': (set(), self._send_playing),
                                'media': (set(), self._send_all_media)}
                self.lock = threading.Lock()
                self.requests_ns = self.refStore.create_namespace()
	def _channel_constructor(self, *args, **kwargs):
		return MariedChannelClass(self, *args, **kwargs)
	def _on_media_changed(self, changeList):
                msg = {'type': 'media_changed'}
                if changeList is None:
                        msg['changes'] = None
                else:
                        msg['changes'] = {
                                'added': [self._get_media_dict(m)
                                        for m in changeList.added],
                                'updated': [self._get_media_dict(m)
                                        for m in changeList.updated],
                                'removed': [self._get_media_dict(m)
                                        for m in changeList.removed]}
                for follower in self._followers_of('media'):
		        follower.send_message(msg)
	def _on_requests_changed(self):
                self._send_all_requests(self._followers_of('requests'))
	def _on_playing_changed(self, previously_playing):
                self._send_playing(self._followers_of('playing'))

        def _get_media_dict(self, media):
                return {'key': str(media.key),
                        'artist': media.artist,
                        'title': media.title,
                        'uploadedByKey': str(media.uploadedByKey),
                        'length': media.length}
	def _send_playing(self, followers):
		playing = self.desk.get_playing()
                msg = { 'type': 'playing',
                        'playing': {
                        'mediaKey': None if playing[0] is None
                                else str(playing[0].key),
                        'byKey': None if playing[1] is None
                                else str(playing[1].byKey),
                        'serverTime': time.time(),
                        'endTime': (time.mktime(playing[2].timetuple()) if
                                not playing[2] is None else None)}}
                for follower in followers:
                        follower.send_message(msg)

        def _send_all_requests(self, followers):
                msg = {
                        'type': 'requests',
                        'requests': [{
                                        'key': self.requests_ns.key_of(r),
                                        'byKey': None if r.byKey is None
                                                else str(r.byKey),
                                        'mediaKey': str(r.mediaKey)
                                     } for r
                                in self.desk.list_requests()]}
                for follower in followers:
                        follower.send_message(msg)

        def _send_all_media(self, followers):
                for follower in followers:
                        follower.send_message({
                                'type': 'media',
                                'count': self.desk.get_media_count()})
                for ms in iter_by_n(self.desk.list_media(), 2):
                        msg = {
                                'type': 'media_part',
                                'part': [self._get_media_dict(m) for m in ms]}
                        for follower in followers:
                                follower.send_message(msg)

        def _followers_of(self, followed):
                with self.lock:
                        return tuple(self.following_lut[followed][0])

        def _register_follower(self, follower, followed):
                with self.lock:
                        if not followed in self.following_lut:
                                raise KeyError
                        self.following_lut[followed][0].add(follower)
                        full_cb = self.following_lut[followed][1]
                full_cb([follower])

        def _unregister_follower(self, follower, followed):
                with self.lock:
                        if not followed in self.following_lut or \
                                        follower not in self.following_lut[
                                                        followed][0]:
                                raise KeyError
                        self.following_lut[followed][0].remove(follower)

        def _remove_follower(self, follower):
                with self.lock:
                        for v in self.following_lut.itervalues():
                                if follower in v[0]:
                                        v[0].remove(follower)




from core import Module
from urllib2 import URLError
import threading
import scrobbler
import time

class Scrobbler(Module):
	def __init__(self, *args, **kwargs):
		super(Scrobbler, self).__init__(*args, **kwargs)
		self.desk.on_playing_changed.register(self._on_playing_changed)
		self.register_on_setting_changed('username', self.osc_creds)
		self.register_on_setting_changed('password', self.osc_creds)
		self.cond = threading.Condition()
		self.running = True
		self.authenticated = False
		self.queue = list()
		self.osc_creds()
	def osc_creds(self):
		if (not hasattr(self, 'username') or
				not hasattr(self, 'password')):
			return
		try:
			scrobbler.login(self.username,
					self.password, hashpw=True)
		except scrobbler.AuthError:
			self.l.error('Couldn\'t authenticate with last.fm')
			self.authenticated = False
			return
		self.authenticated = True

	def _on_playing_changed(self, previous_playing):
		with self.cond:
			self.queue.append(previous_playing)
			self.cond.notify()
	def scrobble(self, media, end_time):
		if not self.authenticated:
			return
		if media.length <= 30:
			self.l.info("%s is too short to be scrobbled" % media)
			return
		time_played = (media.length + time.time() -
				time.mktime(end_time.timetuple()))
		if time_played < 240 and time_played < media.length * 0.5:
			self.l.info("%s has not played long enough" % media)
			return
		scrobbler.submit(media.artist, media.title,
			int(time.mktime(end_time.timetuple()) - media.length),
			length=int(media.length))
		scrobbler.flush()
	def run(self):
		self.cond.acquire()
		while self.running:
			playing = self.desk.get_playing()[0]
			self.cond.release()
			try:
				scrobbler.now_playing(playing.artist,
						      playing.title,
						      length=int(playing.length))
			except URLError as e:
				self.l.exception("Error while scrobbler.now_playing")
			self.cond.acquire()
			while len(self.queue) > 0:
				m, r, end_time = self.queue.pop()
				if not r is None:
					self.cond.release()
					try:
						self.scrobble(m, end_time)
					except URLError as e:
						self.l.exception("Error "+
							"while scrobbler.submit")
					self.cond.acquire()
			if not self.running:
				break
			self.cond.wait()
		self.cond.release()
	def stop(self):
		self.running = False
		with self.cond:
			self.cond.notify()



from __future__ import with_statement
import os
import time
import random
import logging
import datetime
import tempfile
import threading
import collections

from mirte.core import Module
from sarah.event import Event
from sarah.dictlike import DictLike

ChangeList = collections.namedtuple('ChangeList', (
                'added', 'updated', 'removed'))

class Denied(Exception):
	pass
class Stopped(Exception):
	pass
class EmptyQueueException(Exception):
	pass
class AlreadyInQueueError(Denied):
	pass
class MaxQueueLengthExceededError(Denied):
	pass
class MaxQueueCountExceededError(Denied):
	pass
class UnderPreShiftLock(Denied):
        pass


class Media(DictLike):
	def __init__(self, coll, data):
		super(Media, self).__init__(data)
		self.self.coll = coll
	def unlink(self):
		self.coll.unlink_media(self)
	def save(self):
		self.coll.save_media(self)
	@property
	def mediaFile(self):
		return self.coll.mediaStore.by_key(self.mediaFileKey)
	def __eq__(self, other):
		return self.key == other.key
	def __ne__(self, other):
		return self.key != other.key


class MediaFile(object):
	def __init__(self, store, key):
		self._key = key
		self.store = store
	def get_named_file(self):
		raise NotImplementedError
	def open(self):
		raise NotImplementedError
	def remove(self):
		self.store.remove(self)
	def __eq__(self, other):
		return self._key == other.key
	def __ne__(self, other):
		return self._key != other.key
	def get_info(self):
		stream = self.open()
		ret = self.store.mediaInfo.get_info(stream)
		if hasattr(stream, 'close'):
			stream.close()
		return ret
	@property
	def key(self):
		return self._key
class BaseRequest(DictLike):
	pass
class PastRequest(BaseRequest):
	def __init__(self, history, data):
		super(PastRequest, self).__init__(data)
		self.self.history = history
	def remove(self):
		raise NotImplementedError
class Request(BaseRequest):
	def __init__(self, queue, data):
		super(Request, self).__init__(data)
		self.self.queue = queue
	def move(self, amount):
		self.queue.move(self, amount)
	def cancel(self):
		self.queue.cancel(self)
        @property
        def mediaKey(self):
                return self.media.key
        @property
        def byKey(self):
                try:
                        return self.by.key
                except AttributeError:
                        return None
class User(DictLike):
	def has_access(self):
		raise NotImplementedError
	def __eq__(self, other):
		return self.key == other.key
	def __ne__(self, other):
		return self.key != other.key

class Desk(Module):
	def __init__(self, *args, **kwargs):
		super(Desk, self).__init__(*args, **kwargs)
		self.on_playing_changed = Event()
		self.on_media_changed = Event()
		self.on_requests_changed = Event()
		self.orchestrator.on_playing_changed.register(
				self._on_playing_changed)
		self.collection.on_changed.register(
				self._on_collection_changed)
                self.queue.on_changed.register(
                                self._on_requests_changed)
	def _on_playing_changed(self, previous_playing):
		self.on_playing_changed(previous_playing)
	def _on_collection_changed(self, changeList):
		self.on_media_changed(changeList)
	def _on_requests_changed(self):
		self.on_requests_changed()
	def list_media(self):
		return self.collection.media
	def list_media_keys(self):
		return self.collection.media_keys
        def get_media_count(self):
                return self.collection.media_count
	def request_media(self, media, user):
		self.users.assert_request(user, media)
		self.queue.request(media, user)
	def add_media(self, stream, user, customInfo=None):
		mediaFile = self.mediaStore.create(stream)
		try:
			self.users.assert_addition(user, mediaFile)
		except Denied:
			mediaFile.remove()
			raise
		self.collection.add(mediaFile, user, customInfo)
	def list_requests(self):
		return self.queue.requests
        def skip_playing(self, user):
                playing = self.orchestrator.get_playing()
                # We check the end time to avoid to skip a race conditon
                # and skip the next song.
                if playing[2]  - datetime.datetime.now() \
                                < datetime.timedelta(0, 1):
                        raise Denied
                self.users.assert_skip(user, playing[1])
                self.orchestrator.skip()
	def cancel_request(self, request, user):
		self.users.assert_cancel(user, request)
		request.cancel()
	def move_request(self, request, amount, user):
		self.users.assert_move(user, request, amount)
		request.move(amount)
	def get_playing(self):
		return self.orchestrator.get_playing()
	def user_by_key(self, key):
		return self.users.by_key(key)
	def media_by_key(self, key):
		return self.collection.by_key(key)

class History(Module):
	def __init__(self, *args, **kwargs):
		super(History, self).__init__(*args, **kwargs)
		# when a record is added runtime
		self.on_record = Event()
		# when the whole history is changed (db change, et al)
		self.on_pretty_changed = Event()
	def record(self, media, request, at):
		raise NotImplementedError
	def list_past_requests(self):
		raise NotImplementedError

class Users(Module):
	def assert_request(self, user, media):
		raise NotImplementedError
	def assert_addition(self, user, mediaFile):
		raise NotImplementedError
	def assert_cancel(self, user, request):
		raise NotImplementedError
	def assert_move(self, user, request, amount):
		raise NotImplementedError
	def assert_access(self, user):
		return NotImplementedError
	def by_key(self, key):
		return NotImplementedError

class RandomQueue(Module):
	def __init__(self, *args, **kwargs):
		super(RandomQueue, self).__init__(*args, **kwargs)
                self.lock = threading.Lock()
		self.list = list()
                self.on_changed = Event()
		self.register_on_setting_changed('length', self.osc_length)
		self.osc_length()
		self.random.on_ready.register(self._random_on_ready)
                self.pre_shift_lock = False
	def _random_on_ready(self):
                with self.lock:
                        self._fill()
                self.on_changed()
	@property
	def requests(self):
                with self.lock:
                        return reversed(self.list)
        def peek(self, set_pre_shift_lock=False):
                with self.lock:
                        if not self.list:
                                raise EmptyQueueException
                        ret = self.list[-1]
                        if set_pre_shift_lock:
                                self.pre_shift_lock = True
                        return ret
	def shift(self):
                with self.lock:
                        if not self.list:
                                raise EmptyQueueException
                        ret = self.list.pop()
                        self._grow()
                        self.pre_shift_lock = False
                self.on_changed()
                return ret
	def _grow(self):
                if self.random.ready:
                        self.list.insert(0, Request(self, {
                                'media': self.random.pick()}))
	def request(self, media, user):
		assert False # shouldn't do that
	def cancel(self, request):
                with self.lock:
                        if self.pre_shift_lock and request == self.list[-1]:
                                raise UnderPreShiftLock
                        self.list.remove(request)
                        self._grow()
                self.on_changed()
	def move(self, request, amount):
		assert False # shouldn't do that
	def osc_length(self):
                with self.lock:
                        self._fill()
                self.on_changed()
	def _fill(self):
		if not self.random.ready:
			return
		if len(self.list) < self.length:
			for i in xrange(self.length - len(self.list)):
				self._grow()
		else:
			self.list = self.list[:self.length]

class AmalgamatedQueue(Module):
	def __init__(self, *args, **kwargs):
		super(AmalgamatedQueue, self).__init__(*args, **kwargs)
                self.on_changed = Event()
                self.register_on_setting_changed('first', self.osc_first)
                self.register_on_setting_changed('second', self.osc_second)
                self.osc_first()
                self.osc_second()
        def peek(self, set_pre_shift_lock):
                try:
                        return self.first.peek(set_pre_shift_lock)
                except EmptyQueueException:
                        return self.second.peek(set_pre_shift_lock)
        def osc_first(self):
                self.first.on_changed.register(self._subqueue_changed)
        def osc_second(self):
                self.second.on_changed.register(self._subqueue_changed)
        def _subqueue_changed(self):
                self.on_changed()
	def request(self, media, user):
		self.first.request(media, user)
	@property
	def requests(self):
		return (tuple(self.first.requests) +
			tuple(self.second.requests))
	def shift(self):
		assert False # shouldn't do that
	def cancel(self, request):
		if request in self.first.requests:
			self.first.cancel(request)
		else:
			self.second.cancel(request)
	def move(self, request, amount):
		if request in self.first.requests:
			self.first.move(request, amount)
		else:
			self.second.move(request, amount)

class Queue(Module):
	def __init__(self, *args, **kwargs):
		super(Queue, self).__init__(*args, **kwargs)
		self.list = list()
		self.lock = threading.Lock()
                self.on_changed = Event()
                self.pre_shift_lock = False
	def request(self, media, user):
		with self.lock:
			self.list.insert(0, Request(self, {
                                        'media': media,
                                        'by': user}))
                self.on_changed()
	@property
	def requests(self):
		with self.lock:
			return reversed(self.list)
        def peek(self, set_pre_shift_lock=False):
                """ Returns the first element in the Queue

                If set_pre_shift_lock is set, then the queue will disallow
                any modification to the first element in the Queue, until
                it has been shifted. """
                with self.lock:
                        if not self.list:
                                raise EmptyQueueException
                        ret = self.list[-1]
                        if set_pre_shift_lock:
                                self.pre_shift_lock = True
                        return ret
	def shift(self):
		with self.lock:
			if not self.list:
				raise EmptyQueueException
			ret = self.list.pop()
                        self.pre_shift_lock = False
                self.on_changed()
                return ret
	def cancel(self, request):
		with self.lock:
                        if self.pre_shift_lock and self.list[-1] == request:
                                raise UnderPreShiftLock
			self.list.remove(request)
                self.on_changed()
	def move(self, request, amount):
		aa = abs(amount)
		with self.lock:
			o = self.list if amount == aa else list(
                                                reversed(self.list))
                        if self.pre_shift_lock and o[-1] == request:
                                raise UnderPreShiftLock
			idx = o.index(request)
			n = (self.list[:idx] + 
			     self.list[idx+1:idx+aa+1] +
			     [self.list[idx]] +
			     self.list[idx+aa+1:])
			self.list = n if amount == aa else list(reversed(n))
                self.on_changed()

class Orchestrator(Module):
	def __init__(self, *args, **kwargs):
		super(Orchestrator, self).__init__(*args, **kwargs)
		self.on_playing_changed = Event()
		self.lock = threading.Lock()
		self.playing_media = None
		self.satisfied_request = None
                self.player.on_about_to_finish.register(
                                self._player_on_about_to_finish)
                self.player.on_playing_started.register(
                                self._player_on_playing_started)
                self.player.on_playing_finished.register(
                                self._player_on_playing_finished)
                self.peeked_from_randomQueue = True
                self.next_satisfied_request = None
                self.next_playing_media = None
                self.previously_playing = None

	def get_playing(self):
		with self.lock:
			return (self.playing_media,
				self.satisfied_request,
				self.player.endTime)
        def run(self):
                with self.lock:
                        self.running = True
                self._queue_next()

	def stop(self):
		with self.lock:
			self.running = False

        def _queue_next(self):
                self.lock.acquire()
                try:
                        if not self.running:
                                return
                        req = None
                        try:
                                req = self.queue.peek(set_pre_shift_lock=True)
                                media = req.media
                                assert not media is None
                                self.peeked_from_randomQueue = False
                        except EmptyQueueException:
                                media = self._peek_from_randomQueue()
                        self.next_playing_media = media
                        self.next_satisfied_request = req
                        self.player.queue(media)
                finally:
                        self.lock.release()

        def skip(self):
                self.player.skip()

        def _player_on_playing_finished(self, media, endTime):
                with self.lock:
                        assert media == self.playing_media
                        self.previously_playing = (media,
                                                 self.satisfied_request,
                                                 endTime)
                if not media is None:
                        self.history.record(media,
                                self.satisfied_request,
                                endTime  - datetime.timedelta(0, media.length))

        def _player_on_playing_started(self, media, endTime):
                with self.lock:
                        assert media == self.next_playing_media
                        self.satisfied_request = self.next_satisfied_request
                        self.next_satisfied_request = None
                        self.next_playing_media = None
                        self.playing_media = media
                        if self.peeked_from_randomQueue:
                                self.randomQueue.shift()
                        else:
                                self.queue.shift()
                        previously_playing = self.previously_playing
                        self.previously_playing = None
                self.on_playing_changed(previously_playing)
        def _peek_from_randomQueue(self):
                while True:
                        try:
                                media = self.randomQueue.peek(
                                        set_pre_shift_lock=True).media
                                self.peeked_from_randomQueue = True
                                return media
                        except EmptyQueueException:
                                self.lock.release()
                                self.wait_for_media()
                                self.lock.acquire()

        def _player_on_about_to_finish(self):
                self._queue_next()
	
	def wait_for_media(self):
		self.l.info("Randomqueue couldn't return media -- collection "+
			    "is assumed to be empty -- waiting for media.")
		self.randomQueue.random.collection.got_media_event.wait()
		self.l.info("Woke!")

class Random(Module):
	def __init__(self, *args, **kwargs):
		super(Random, self).__init__(*args, **kwargs)
		self.cond = threading.Condition()
		self.running = True
		# used to push new PastRequest s to the worker thread
		# None is used in the case of "history.on_pretty_changed"
		self.recordStack = list()
		self.on_ready = Event()
		self.history.on_pretty_changed.register(
				self._on_history_pretty_changed)
		self.history.on_record.register(
				self._on_history_record)

	def _on_history_pretty_changed(self):
		with self.cond:
			self.recordStack = [None]
			self.cond.notify()
	def _on_history_record(self, pr):
		with self.cond:
			self.recordStack.append(pr)
			self.cond.notify()

	def pick(self):
		raise NotImplementedError

	def run(self):
		while True:
			with self.cond:
				self.cond.wait()
			if not self.running:
				break
			if len(self.recordStack) == 0:
				continue
			for pr in reversed(self.recordStack):
				if pr is None:
					self._handle_history_pretty_changed()
				else:
					self._handle_history_record(pr)

	def stop(self):
		self.running = False
		with self.cond:
			self.cond.notify()
	
	def _handle_history_record(self, pr):
		raise NotImplementedError
	def _handle_history_pretty_changed(self):
		raise NotImplementedError

class SimpleRandom(Random):
	def __init__(self, *args, **kwargs):
		super(SimpleRandom, self).__init__(*args, **kwargs)
		self.collection.on_keys_changed.register(
				self.on_collection_keys_changed)
		self.on_collection_keys_changed(None)
	
	def on_collection_keys_changed(self, changeList):
                if changeList is None:
                        self.keys = self.collection.media_keys
                else:
                        for x in changeList.removed:
                                self.keys.remove(x)
                        for x in changeList.added:
                                self.keys.append(x)
		if len(self.keys) > 0:
			self.on_ready()

	def pick(self):
		if len(self.keys) == 0:
			return None
		key = self.keys[random.randint(0, len(self.keys) - 1)]
		return self.collection.by_key(key)

	def _handle_history_record(self, pr):
		pass
	def _handle_history_pretty_changed(self):
		pass
	@property
	def ready(self):
		return len(self.keys) > 0

class MediaInfo(Module):
	def get_info(self, stream):
		pass
	def get_info_by_path(self, path):
		with open(path) as f:
			return self.get_info(f)

class MediaStore(Module):
	def create(self, stream):
		raise NotImplementedError
	def by_key(self, key):
		raise NotImplementedError
	@property
	def keys(self):
		raise NotImplementedError

class Player(Module):
	def __init__(self, *args, **kwargs):
		super(Player, self).__init__(*args, **kwargs)
		self.endTime = None
                self.on_about_to_finish = Event()
                self.on_playing_started = Event()
                self.on_playing_finished = Event()
	def stop(self):
		raise NotImplementedError
	def queue(self, media):
		raise NotImplementedError

class Collection(Module):
	def __init__(self, *args, **kwargs):
		super(Collection, self).__init__(*args, **kwargs)
		self.on_keys_changed = Event()
		self.on_changed = Event()
		# got_media_event is set when the Collection isn't
		# empty.
		self.got_media_event = threading.Event()
	def add(self, mediaFile, user, extraInfo=None):
		raise NotImplementedError
	@property
	def media(self):
		raise NotImplementedError
	@property
	def media_keys(self):
		raise NotImplementedError
	def by_key(self, key):
		raise NotImplementedError
	def save_media(self, media):
		raise NotImplementedError
	def unlink_media(self, media):
		raise NotImplementedError




from __future__ import with_statement

from maried.core import MediaStore, MediaFile, Media, Collection, User, Users, \
			History, Request, PastRequest, Denied, \
                        AlreadyInQueueError, MaxQueueLengthExceededError, \
                        ChangeList
from mirte.core import Module
from sarah.event import Event
from sarah.dictlike import AliasingMixin

from pymongo.objectid import ObjectId

import threading
import tempfile
import pymongo
import hashlib
import gridfs
import base64
import time
import os

class MongoMediaFile(MediaFile):
	def open(self):
		return self.store._open(self.key)
	def get_named_file(self):
		return self.store._get_named_file(self.key)
	def __repr__(self):
		return "<MongoMediaFile %s>" % self._key

class MongoUser(AliasingMixin, User):
	aliases = {'key': '_id',
		   'realName': 'n',
		   'level': 'l',
                   'accessKey': 'a',
		   'passwordHash': 'p'}
	def __init__(self, coll, data):
		super(MongoUser, self).__init__(self.normalize_dict(data))
		self.self.collection = coll
	@property
        def has_access(self):
                return self.level >= 2
        @property
        def may_cancel(self):
                return self.level >= 3
        @property
        def may_skip(self):
                return self.level >= 5
        @property
        def is_admin(self):
                return self.level >= 5
        @property
        def may_move(self):
                return self.level >= 3
	def check_password(self, password):
		return self.passwordHash == hashlib.md5(password).hexdigest()
        def set_password(self, password):
                self.passwordHash = hashlib.md5(password).hexdigest()
        def regenerate_accessKey(self):
                self.accessKey = base64.b64encode(os.urandom(6))
        def save(self):
                self.collection._save_user(self)

class MongoMedia(AliasingMixin, Media):
	aliases = {'key': '_id',
		   'artist': 'a',
		   'title': 't',
		   'trackGain': 'tg',
		   'trackPeak': 'tp',
		   'length': 'l',
		   'mediaFileKey': 'k',
		   'uploadedByKey': 'ub',
		   'uploadedTimestamp': 'ut'}
	def __init__(self, coll, data):
		super(MongoMedia, self).__init__(coll,
                                self.normalize_dict(data))
        def __repr__(self):
                return '<MongoMedia %s - %s (%s)>' % (self.artist, self.title,
                                                        str(self.key))

class MongoPastRequest(AliasingMixin, PastRequest):
	aliases = {'key': '_id',
		   'byKey': 'b',
		   'at': 'a',
		   'mediaKey': 'm'}
	def __init__(self, history, data):
		super(MongoPastRequest, self).__init__(history,
                                self.normalize_dict(data))

class MongoRequest(AliasingMixin, PastRequest):
	aliases = {'key': '_id',
		   'byKey': 'b',
		   'mediaKey': 'm'}
	def __init__(self, queue, data):
		super(MongoRequest, self).__init__(queue,
                                self.normalize_dict(data))


class MongoDb(Module):
	def __init__(self, *args, **kwargs):
		super(MongoDb, self).__init__(*args, **kwargs)
		self.register_on_setting_changed('host', self.osc_creds)
		self.register_on_setting_changed('port', self.osc_creds)
		self.register_on_setting_changed('db', self.osc_creds)
		self.on_changed = Event()
		self.ready = None
		self.osc_creds()

	def osc_creds(self):
		if None in (self.host, self.port, self.db):
			return
		self.con = pymongo.Connection(self.host, self.port)
		self.db = self.con[self.db]
		self.ready = True
		self.on_changed()

class MongoCollection(Collection):
	def __init__(self, *args, **kwargs):
		super(MongoCollection, self).__init__(*args, **kwargs)
		self._media = None
		self.lock = threading.Lock()
		self.register_on_setting_changed('db', self.osc_db)
		self.osc_db()
	
	def osc_db(self):
		self.db.on_changed.register(self.on_db_changed)
		self.on_db_changed()

	def on_db_changed(self):
		if not self.db.ready:
			return
		with self.lock:
			self.cMedia = self.db.db['media']
			self.cUsers = self.db.db['users']
			self._media = {}
			self.users = {}
			for tmp in self.cMedia.find():
				self._media[tmp['_id']] = MongoMedia(self, tmp)
			for tmp in self.cUsers.find():
				self.users[tmp['_id']] = MongoUser(self, tmp)
		self.l.info("Cached %s media %s users" % (len(self._media),
							  len(self.users)))
		self.on_keys_changed(None)
		self.on_changed(None)
		with self.lock:
			if len(self._media) > 0:
				self.got_media_event.set()
			else:
				self.got_media_event.clear()
	
	@property
	def media(self):
		if self._media is None:
			return list()
		return self._media.itervalues()

	@property
	def media_keys(self):
		if self._media is None:
			return list()
		return self._media.keys()

        @property
        def media_count(self):
                return len(self._media)

	def by_key(self, key):
                if isinstance(key, basestring):
                        key = ObjectId(key)
		return self._media[key]

	def _user_by_key(self, key):
		return self.users[key]

	def stop(self):
		self.got_media_event.set()

	def add(self, mediaFile, user, extraInfo=None):
		info = mediaFile.get_info()
		if not extraInfo is None:
			info.update(extraInfo)
		info.update({
			'mediaFileKey': mediaFile.key,
			'uploadedTimestamp': time.time(),
			'uploadedByKey': user.key})
		key = self.cMedia.insert(MongoMedia.normalize_dict(info))
		info['_id'] = key
		with self.lock:
			self._media[key] = MongoMedia(self, info)
			if len(self._media) == 1:
				self.got_media_event.set()
		self.on_keys_changed(ChangeList(
                        added=(key,), updated=(), removed=()))
		self.on_changed(ChangeList(
                        added=(self._media[key],), updated=(), removed=()))
	
	def _unlink_media(self, media):
		with self.lock:
			del(self._media[media.key])
		self.cMedia.remove({'_id': media.key})
		self.db.on_keys_changed(added=(), updated=(),
                                removed=(media.key,))
		self.on_changed(added=(), updated=(), removed=(media,))
	
	def _save_media(self, media):
		self.cMedia.save(media.to_dict())
		self.on_changed(added=(), updated=(media,), removed=())

        def _save_user(self, user):
                self.cUsers.save(user.to_dict())

class MongoMediaStore(MediaStore):
	def __init__(self, *args, **kwargs):
		super(MongoMediaStore, self).__init__(*args, **kwargs)
		self.ready = False
		self.keysCond = threading.Condition()
		self._keys = None
		self.register_on_setting_changed('db', self.osc_db)
                self.osc_db()

	def osc_db(self):
		self.db.on_changed.register(self.on_db_changed)
		self.on_db_changed()

	def on_db_changed(self):
		if not self.db.ready:
			return
		self.fs = gridfs.GridFS(self.db.db, self.collection)
		self.threadPool.execute_named(self._do_refresh_keys,
				'%s _do_refresh_keys' % self.l.name)
	
	def _do_refresh_keys(self):
		with self.keysCond:
			self._keys = self.fs.list()
			self.l.info("Got %s files" % len(self._keys))
			self.keysCond.notifyAll()

	def create(self, stream):
		(fd, fn) = tempfile.mkstemp()
		f = open(fn, 'w')
		m = hashlib.sha512()
		while True:
			b = stream.read(2048)
			m.update(b)
			if len(b) == 0:
				break
			f.write(b)
                stream.close()
		f.close()
		hd = m.hexdigest()
		if not self._keys is None and hd in self._keys:
			self.l.warn("Duplicate file %s" % hd)
		else:
			with open(fn) as f:
				self.fs.put(f, filename=hd)
			os.unlink(fn)
			with self.keysCond:
				if self._keys is None:
					self.l.debug(
						"create: waiting on keysCond")
					self.keysCond.wait()
				self._keys.append(hd)
		return self.by_key(hd)

	def by_key(self, key):
		with self.keysCond:
			if self._keys is None:
				self.l.debug("by_key: waiting on keysCond")
				self.keysCond.wait()
			if not key in self._keys:
				raise KeyError, key
		return MongoMediaFile(self, key)

	def remove(self, mediaFile):
		self.l.info("Removing %s" % mediaFile)
		self.fs.delete(mediaFile.key)
	
	def _open(self, key):
		return self.fs.get_last_version(key)

	def _get_named_file(self, key):
		raise NotImplementedError
	
	@property
	def keys(self):
		with self.keysCond:
			if self._keys is None:
				self.l.debug("keys: waiting on keysCond")
				self.keysCond.wait()
			return tuple(self._keys)

class MongoHistory(History):
	def __init__(self, *args, **kwargs):
		super(MongoHistory, self).__init__(*args, **kwargs)
		self.register_on_setting_changed('db', self.osc_db)
		self.osc_db()

        def osc_db(self):
		self.db.on_changed.register(self._on_db_changed)
                self._on_db_changed()
	
	def _on_db_changed(self):
		if not self.db.ready:
			return
		self.cHistory = self.db.db['history']
		self.on_pretty_changed()
	
	def record(self, media, request, at):
		self.l.info(repr(media if request is None else request))
		info = {'mediaKey': media.key,
			'byKey': None if request is None else request.by.key,
			'at':  time.mktime(at.timetuple())}
		info['key'] = self.cHistory.insert(
                                MongoPastRequest.normalize_dict(info))
		self.on_record(MongoPastRequest(self, info))
	
	def list_past_requests(self):
		for tmp in self.cHistory.find():
			yield MongoPastRequest(self, tmp)

class MongoUsers(Users):
        def assert_request(self, user, media):
                if not user.has_access:
                        raise Denied
                requests = self.queue.requests
                if any(map(lambda x: x.media == media, requests)):
                        raise AlreadyInQueueError
                ureqs = filter(lambda y: y.by == user, requests)
                if len(ureqs) > self.maxQueueCount:
                        raise MaxQueueCountExceededError
                if (sum(map(lambda x: x.media.length, ureqs)) >
                                self.maxQueueLength):
                        raise MaxQueueLengthExceededError
        def assert_addition(self, user, mediaFile):
                pass
        def assert_cancel(self, user, request):
                if user.is_admin:
                        return
                if request.by == user:
                        return
                if not user.may_cancel:
                        raise Denied
        def assert_move(self, user, request, amount):
                if request.by == user and amount < 0:
                        return
                if not user.may_move:
                        raise Denied
        def assert_skip(self, user, request):
                if not user.may_skip:
                        raise Denied
        def by_key(self, key):
                return self.collection._user_by_key(key)



#!/usr/bin/env python

from setuptools import setup
from get_git_version import get_git_version

setup(name='maried',
      version=get_git_version(),
      description='MarieD music daemon',
      author='Bas Westerbaan',
      author_email='bas@westerbaan.name',
      url='http://github.com/bwesterb/maried/',
      packages=['maried'],
      package_data={'': ['*.mirte']},
      zip_safe=False,
      package_dir={'maried': 'src'},
      install_requires = ['docutils>=0.3',
			  'mirte>=0.1.0a2',
			  'sarah>=0.1.0a2',
                          'joyce>=0.1.0a2'],
      )



from __future__ import with_statement

# gtk, gst, gobject and pygst are imported in GtkMainLoop.run
gtk, gst, gobject, pygst = None, None, None, None

import time
import os.path
import datetime
import threading

from maried.core import Player, MediaInfo, Stopped
from mirte.core import Module
from sarah.event import Event
from sarah.io import SocketPairWrappedFile

class GtkMainLoop(Module):
        def __init__(self, *args, **kwargs):
                super(GtkMainLoop, self).__init__(*args, **kwargs)
                self.ready = threading.Event()
	def run(self):
                self.do_imports()
		gtk.gdk.threads_init()
                self.ready.set()
		gtk.main()
        def do_imports(self):
                global gtk, gst, gobject, pygst
                import gtk as _gtk
                import gst as _gst
                import gobject as _gobject
                import pygst as _pygst
                gtk, gst, gobject, pygst = _gtk, _gst, _gobject, _pygst
	def stop(self):
		gtk.main_quit()

class GstMediaInfo(MediaInfo):
	class Job(object):
		def __init__(self, mi, uri):
			self.mi = mi
			self.event = threading.Event()
			self.inError = False
			self.result = dict()
			self.bin = gst.element_factory_make(
					'playbin', 'playbin')
			fakesink = gst.element_factory_make(
					'fakesink', 'fakesink')
			rganalysis = gst.element_factory_make(
					'rganalysis', 'rganalysis')
			bin2 = gst.element_factory_make('bin', 'bin')
			bin2.add(rganalysis)
			bin2.add(fakesink)
			rganalysis.link(fakesink)
			bin2.add_pad(gst.GhostPad('ghostpad',
					rganalysis.get_static_pad('sink')))
			self.bin.set_property('video-sink', fakesink)
			self.bin.set_property('audio-sink', bin2)
			self.bus = bus = self.bin.get_bus()
			bus.add_signal_watch()
			bus.connect('message', self.on_message)
			bus.connect('message::tag', self.on_tag)
			self.bin.set_property('uri', uri)
			self.bin.set_state(gst.STATE_PLAYING)
		
		def on_message(self, bus, message):
			t = message.type
			if t == gst.MESSAGE_ERROR:
				error, debug = message.parse_error()
				self.mi.l.error("Gst: %s %s" % (error, debug))
				self.inError = True
				self.finish()
			elif t == gst.MESSAGE_EOS:
				self.on_eos()

		def on_eos(self):
			rawpos = self.bin.query_position(
					gst.FORMAT_TIME)[0]
			try:
				rawdur = self.bin.query_duration(
					gst.FORMAT_TIME)[0]
			except gst.QueryError:
				rawdur = -1
			if rawdur == -1:	
				self.mi.l.warn('query_duration failed, '+
					'falling back to query_position')
				raw = rawpos
			else:
				raw = rawdur
			self.result['length'] = raw / (1000.0**3)
			self.finish()
		
		def on_tag(self, bus, message):
			tagList = message.parse_tag()
			for key in tagList.keys():
				if key == 'artist':
					self.result[key] = tagList[key]
				elif key == 'title':
					self.result[key] = tagList[key]
				elif key == gst.TAG_TRACK_PEAK:
					self.result['trackPeak'] = tagList[key]
				elif key == gst.TAG_TRACK_GAIN:
					self.result['trackGain'] = tagList[key]

		def interrupt(self):
			self.inError = True
			self.finish()

		def finish(self):
			self.event.set()
			self.bin.set_state(gst.STATE_NULL)
			self.bus.remove_signal_watch()
			del(self.bin)
			del(self.bus)
	
	def __init__(self, *args, **kwargs):
		super(GstMediaInfo, self).__init__(*args, **kwargs)
		self.lock = threading.Lock()
		self.jobs = set()
	
	def get_info(self, stream):
		wrapper = None
		if hasattr(stream, 'fileno'):
			uri = 'fd://%s' % stream.fileno()
		else:
			wrapper = SocketPairWrappedFile(stream)
			uri = 'fd://%s' % wrapper.fileno()
			self.threadPool.execute_named(wrapper.run,
					'%s wrapper.run %s' % (self.l.name,
						wrapper.fileno()))
		j = GstMediaInfo.Job(self, uri)
		with self.lock:
			self.jobs.add(j)
		j.event.wait()
		with self.lock:
			self.jobs.remove(j)
		if not wrapper is None:
			wrapper.close()
		if j.inError:
			raise ValueError
		return j.result

	def get_info_by_path(self, path):
		uri = 'file://' + os.path.abspath(path)
		j = GstMediaInfo.Job(self, uri)
		with self.lock:
			self.jobs.add(j)
		j.event.wait()
		with self.lock:
			self.jobs.remove(j)
		if j.inError:
			raise ValueError
		return j.result

class GstPlayer(Player):
	def __init__(self, *args, **kwargs):
		super(GstPlayer, self).__init__(*args, **kwargs)
                self.readyEvent = threading.Event()
		self.idleCond = threading.Condition()

        def run(self):
                self._initialize()

        def _initialize(self):
                self.gtkMainLoop.ready.wait()
                
                # set up the player bin
		self.bin = gst.element_factory_make('playbin2', 'playbin2')
		self.bin2 = gst.element_factory_make('bin', 'bin')
		self.ac = gst.element_factory_make('audioconvert',
				'audioconvert')
		self.ac2 = gst.element_factory_make('audioconvert',
				'audioconvert2')
		self.rgvolume = gst.element_factory_make('rgvolume', 'rgvolume')
		self.autoaudiosink = gst.element_factory_make('autoaudiosink',
					'autoaudiosink')
		self.bin2.add(self.rgvolume)
		self.bin2.add(self.ac)
		self.bin2.add(self.ac2)
		self.bin2.add(self.autoaudiosink)
		self.ac.link(self.rgvolume)
		self.rgvolume.link(self.ac2)
		self.ac2.link(self.autoaudiosink)
		self.bin2.add_pad(gst.GhostPad('ghostpad',
			self.ac.get_static_pad('sink')))
		self.bin.set_property('audio-sink', self.bin2)
		self.bus = self.bin.get_bus()
		self.bus.add_signal_watch()
		self.bus.connect('message', self.on_message)
                self.bin.connect('about-to-finish', self._on_about_to_finish)

                # internal state
		self.idle = True
		self.stopped = False
                self.playing_media = None
                self.next_media = None
                self.next_wrapper = None
                self.next_stream = None
                self.previous_wrapper = None
                self.previous_stream = None
                self.in_a_skip = False
                self.readyEvent.set()

	def queue(self, media):
                """ Queues the next media to be played """
                # Check current state
                self.readyEvent.wait()
                got_to_start_playing = False
		with self.idleCond:
			if self.stopped:
				raise Stopped
                        if not self.next_media is None:
                                raise RuntimeError, \
                                        "Another media has been queued already"
                        if self.playing_media is None:
                                got_to_start_playing = True
                        self.next_media = media

                # Get mediafile and set up file descriptor
                self.l.debug('queueing %s' % media)
		try:
			mf = media.mediaFile
		except KeyError:
			self.l.error("%s's mediafile doesn't exist" % media)
			return
		stream = mf.open()
		wrapper = None
		if hasattr(stream, 'fileno'):
			uri = 'fd://%s'%stream.fileno()
		else:
			wrapper = SocketPairWrappedFile(stream)
			self.threadPool.execute_named(wrapper.run,
				'%s wrapper.run %s' % (self.l.name,
					wrapper.fileno()))
			uri = 'fd://%s'%wrapper.fileno()
                self.next_wrapper = wrapper
                self.next_stream = stream

                # Queue in the playbin
		self.bin.set_property('uri', uri)

                # Inject TRACK_GAIN and TRACK_PEAK tags
		tl = gst.TagList()
		tl[gst.TAG_TRACK_GAIN] = media.trackGain
		tl[gst.TAG_TRACK_PEAK] = media.trackPeak
		self.rg_event = gst.event_new_tag(tl)

                # Start playing -- if not already
                if got_to_start_playing or self.in_a_skip:
                        self.bin.set_state(gst.STATE_PLAYING)
                        self.idle = False

        def _on_stream_changed(self):
                """ Called when GStreamer signals that the playing stream
                    has changed.  In this method we will clean up the previous
                    stream (if any) and let the world know the playing
                    media changed """
                now = time.time()
                
                self.l.info("playing: %s"% self.next_media)
                
                self._on_media_finished()

                # Update state
                old_endTime = self.endTime
                old_media = self.playing_media
                self.playing_media, self.next_media = self.next_media, None
		self.endTime = datetime.datetime.fromtimestamp(
				now + self.playing_media.length)
                self.previous_stream, self.next_stream = self.next_stream, None
                self.previous_wrapper, self.next_wrapper = \
                                                self.next_wrapper, None
                if self.in_a_skip:
                        self.in_a_skip = False

                # Notify the world
                if not old_media is None:
                        self.on_playing_finished(old_media, old_endTime)
                self.on_playing_started(self.playing_media, self.endTime)

        def _on_about_to_finish(self, bin):
                self.l.info('about to finish')
                self.on_about_to_finish()

	def _on_eos(self):
		with self.idleCond:
                        if not self.next_media is None:
                                return
                        self.bin.set_state(gst.STATE_NULL)
                        self._on_media_finished()
                        self.idle = True
                        self.idleCond.notifyAll()

        def _on_media_finished(self):
                if not self.previous_wrapper is None:
                        self.previous_wrapper.close()
                if hasattr(self.previous_stream, 'close'):
                        self.previous_stream.close()

                # Update state
                old_playing = self.playing_media
                old_endTime = self.endTime
                self.playing_media = None
                self.previous_stream = None
                self.previous_wrapper = None
                self.endTime = None

                # Notify the world
                if not old_playing is None:
                        self.on_playing_finished(old_playing, old_endTime)

        def skip(self):
                with self.idleCond:
                        if self.in_a_skip:
                                raise RuntimeError, "Already skipping"
                        self.l.debug('Skipping')
                        self.in_a_skip = True
                        self.bin.set_state(gst.STATE_NULL)
                        self.on_about_to_finish()

        def _interrupt(self):
                with self.idleCond:
                        self.l.debug('_interrupt!')
                        self.bin.set_state(gst.STATE_NULL)
                        self._on_media_finished()
                        self.idle = True
                        self.idleCond.notifyAll()
	
	def on_message(self, bus, message):
		if message.type == gst.MESSAGE_ERROR:
			error, debug = message.parse_error()
			self.l.error("Gst: %s %s" % (error, debug))
			self._on_eos()
		elif (not self.rg_event is None and
				message.type == gst.MESSAGE_STATE_CHANGED and
				message.src == self.rgvolume and
				message.parse_state_changed()[1] ==
					gst.STATE_PAUSED):
			self.ac.get_static_pad('src').push_event(
					self.rg_event)
			self.rg_event = None
			tg = self.rgvolume.get_property('target-gain')
			rg = self.rgvolume.get_property('result-gain')
			if tg != rg:
				self.l.warn('replaygain: target gain '+
					'not reached: trg %s res %s' % (
						tg, rg))
		elif message.type == gst.MESSAGE_EOS:
			self._on_eos()
                elif (message.type == gst.MESSAGE_ELEMENT and
                                message.structure.get_name() ==
                                        'playbin2-stream-changed'):
                        self._on_stream_changed()

	
	def stop(self):
                self.readyEvent.wait()
                self._interrupt()
		self.bus.remove_signal_watch()
		with self.idleCond:
			if not self.idle:
				self.idleCond.wait()
			self.stopped = True
		del(self.bin)
		del(self.bus)



import time

def sanitize_collection_info(collection, fix=False):
	for media in collection.media:
		info = media.mediaFile.get_info()
		ok = True
		if info['length'] != media.length:
			print "wrong length: %s != %s for %s" % (
					info['length'],
					media.length, media)
			if fix:
				media.length = info['length']
			ok = False
		if (info['track_peak'], info['trackGain']) != (
				media.trackPeak, media.trackGain):
			print "wrong gain/peak: %s != %s for %s" % (
				(info['trackGain'], info['trackPeak']),
				(media.trackPeak, media.trackGain), media)
			if fix:
				media.trackGain = info['trackGain']
				media.trackPeak = info['trackPeak']
			ok = False
		if fix and not ok:
			media.save()
